from common.serializers.json_serializer import JsonSerializer
from indy_common.authorize.auth_request_validator import WriteRequestValidator

from indy_common.constants import RICH_SCHEMA_MAPPING, RS_MAPPING_SCHEMA, RS_CONTENT, RS_TYPE, RS_SCHEMA_TYPE_VALUE, \
    RS_MAPPING_ENC, RS_MAPPING_RANK, RS_ENCODING_TYPE_VALUE, RS_MAPPING_ATTRIBUTES, RS_MAPPING_ISSUER, \
    RS_MAPPING_ISSUANCE_DATE

from indy_node.server.request_handlers.domain_req_handlers.rich_schema.abstract_rich_schema_object_handler import \
    AbstractRichSchemaObjectHandler
from plenum.common.exceptions import InvalidClientRequest

from plenum.server.database_manager import DatabaseManager
from stp_core.common.log import getlogger

logger = getlogger()


class RichSchemaMappingHandler(AbstractRichSchemaObjectHandler):

    def __init__(self, database_manager: DatabaseManager,
                 write_req_validator: WriteRequestValidator):
        super().__init__(RICH_SCHEMA_MAPPING, database_manager, write_req_validator)

    def is_json_ld_content(self):
        return True

    def do_static_validation_content(self, content_as_dict, request):
        # 1. check for missing `schema` or `attributes` fields
        missing_fields = []
        for field in [RS_MAPPING_SCHEMA, RS_MAPPING_ATTRIBUTES]:
            if not content_as_dict.get(field):
                missing_fields.append("'{}'".format(field))

        if missing_fields:
            missing_fields_str = " and ".join(missing_fields)
            raise InvalidClientRequest(request.identifier, request.reqId,
                                       "{} must be set in '{}'".format(missing_fields_str, RS_CONTENT))

        # 2. check for missing defaukt attributes in `attributes` (`issuer` and `issuanceDate`)
        missing_fields = []
        for field in [RS_MAPPING_ISSUER, RS_MAPPING_ISSUANCE_DATE]:
            if not content_as_dict[RS_MAPPING_ATTRIBUTES].get(field):
                missing_fields.append("'{}'".format(field))

        if missing_fields:
            missing_fields_str = " and ".join(missing_fields)
            raise InvalidClientRequest(request.identifier, request.reqId,
                                       "{} must be in {}'s '{}'".format(missing_fields_str, RS_CONTENT,
                                                                        RS_MAPPING_ATTRIBUTES))

    def do_dynamic_validation_content(self, request):
        # it has been checked on static validation step that the content is a valid JSON.
        # and it has schema and attributes fields
        content_as_dict = JsonSerializer.loads(request.operation[RS_CONTENT])

        # 1. check that the schema field points to an existing object on the ledger
        schema_id = content_as_dict[RS_MAPPING_SCHEMA]
        schema, _, _ = self.get_from_state(schema_id)
        if not schema:
            raise InvalidClientRequest(request.identifier,
                                       request.reqId,
                                       'Can not find a schema with id={}; please make sure that it has been added to the ledger'.format(
                                           schema_id))

        # 2. check that the schema field points to an object of the Schema type
        if schema.get(RS_TYPE) != RS_SCHEMA_TYPE_VALUE:
            raise InvalidClientRequest(request.identifier,
                                       request.reqId,
                                       "'{}' field must reference a schema with {}={}".format(
                                           RS_MAPPING_SCHEMA, RS_TYPE, RS_SCHEMA_TYPE_VALUE))

        # 3. find all attribute leaf dicts with encoding-rank pairs
        enc_desc_dicts = list(find_encoding_desc_dicts(content_as_dict[RS_MAPPING_ATTRIBUTES]))

        # 4. check that every dict has encoding and rank fields
        # Note: this check can be done in static validation, but then we will have to traverse the leaf dicts twice
        for desc_dict, attr in enc_desc_dicts:
            if not isinstance(desc_dict, dict):
                raise InvalidClientRequest(request.identifier,
                                           request.reqId,
                                           "{} and {} must be set for the attribute '{}'".format(RS_MAPPING_ENC,
                                                                                                 RS_MAPPING_RANK, attr))

            missing_fields = []
            for field in [RS_MAPPING_ENC, RS_MAPPING_RANK]:
                v = desc_dict.get(field)
                if not v and v != 0:
                    missing_fields.append(field)

            if missing_fields:
                missing_fields_str = " and ".join(missing_fields)
                raise InvalidClientRequest(request.identifier, request.reqId,
                                           "{} must be set for the attribute '{}'".format(missing_fields_str, attr))

        # 5. check that all ranks are unique and form a sequence without gaps
        # Note: this check can be done in static validation, but then we will have to traverse the leaf dicts twice
        expected_ranks = list(range(1, len(enc_desc_dicts) + 1))
        ranks = sorted([desc_dict[RS_MAPPING_RANK] for desc_dict, attr in enc_desc_dicts])
        if ranks != expected_ranks:
            raise InvalidClientRequest(request.identifier, request.reqId,
                                       "the attribute's ranks are not sequential: expected ranks are all values from 1 to {}".format(
                                           len(enc_desc_dicts)))

        # 6. check that all the enc fields point to an existing object on the ledger of the type Encoding
        for desc_dict, attr in enc_desc_dicts:
            encoding_id = desc_dict[RS_MAPPING_ENC]
            encoding, _, _ = self.get_from_state(encoding_id)
            if not encoding:
                raise InvalidClientRequest(request.identifier,
                                           request.reqId,
                                           "Can not find a referenced '{}' with id={} in the '{}' attribute; please make sure that it has been added to the ledger".format(
                                               RS_MAPPING_ENC, encoding_id, attr))
            if encoding.get(RS_TYPE) != RS_ENCODING_TYPE_VALUE:
                raise InvalidClientRequest(request.identifier,
                                           request.reqId,
                                           "'{}' field in the '{}' attribute must reference an encoding with {}={}".format(
                                               RS_MAPPING_ENC, attr, RS_TYPE, RS_ENCODING_TYPE_VALUE))


def find_encoding_desc_dicts(content, last_attr=None):
    if is_leaf_dict(content):
        yield content, last_attr
        return

    for k, v in content.items():
        if k in []:
            continue
        if isinstance(v, list):
            for item in v:
                for desc_dict, last_attr in find_encoding_desc_dicts(item, k):
                    yield desc_dict, last_attr
        else:
            for desc_dict, last_attr in find_encoding_desc_dicts(v, k):
                yield desc_dict, last_attr


def is_leaf_dict(node):
    if isinstance(node, dict):
        for k, v in node.items():
            if isinstance(v, dict):
                return False
            if isinstance(v, list):
                return False
    return True



import json
import copy
from indy_common.constants import CRED_DEF_ID, ID, REVOC_TYPE, TAG, GET_REVOC_REG_DEF, \
    TXN_TYPE, TIMESTAMP, REVOC_REG_DEF_ID, VALUE, FROM, TO, ISSUED, \
    REVOKED, PREV_ACCUM, ACCUM
from plenum.common.constants import STATE_PROOF
from indy_common.state import domain
from plenum.common.txn_util import get_txn_time
from plenum.common.util import randomString
from indy_node.test.anon_creds.helper import check_valid_proof, build_get_revoc_reg_delta, build_get_revoc_reg_entry
from plenum.test.helper import sdk_sign_request_from_dict
from plenum.test.helper import sdk_send_and_check
from plenum.common.util import get_utc_epoch


def test_state_proof_returned_for_get_revoc_reg_def(looper,
                                                    txnPoolNodeSet,
                                                    sdk_wallet_steward,
                                                    sdk_pool_handle,
                                                    send_revoc_reg_def_by_default):
    _, author_did = sdk_wallet_steward
    revoc_req, _ = send_revoc_reg_def_by_default
    get_revoc_reg_def_req = {
        ID: ":".join([author_did,
                      domain.MARKER_REVOC_DEF,
                      revoc_req['operation'][CRED_DEF_ID],
                      revoc_req['operation'][REVOC_TYPE],
                      revoc_req['operation'][TAG]]),
        TXN_TYPE: GET_REVOC_REG_DEF,
    }
    get_revoc_reg_def_req = sdk_sign_request_from_dict(looper,
                                                       sdk_wallet_steward,
                                                       get_revoc_reg_def_req)
    sdk_reply = sdk_send_and_check([json.dumps(get_revoc_reg_def_req)],
                                 looper,
                                 txnPoolNodeSet,
                                 sdk_pool_handle)
    reply = sdk_reply[0][1]
    check_valid_proof(reply)


def test_state_proof_returned_for_get_revoc_reg(looper,
                                                txnPoolNodeSet,
                                                sdk_pool_handle,
                                                send_revoc_reg_entry_by_default,
                                                sdk_wallet_steward):
    rev_entry_req, reg_reply = send_revoc_reg_entry_by_default
    get_revoc_reg = build_get_revoc_reg_entry(looper, sdk_wallet_steward)
    get_revoc_reg['operation'][REVOC_REG_DEF_ID] = rev_entry_req['operation'][REVOC_REG_DEF_ID]
    get_revoc_reg['operation'][TIMESTAMP] = get_utc_epoch() + 1000
    sdk_reply = sdk_send_and_check([json.dumps(get_revoc_reg)], looper, txnPoolNodeSet, sdk_pool_handle)
    reply = sdk_reply[0][1]
    check_valid_proof(reply)


def test_state_proof_returned_for_get_revoc_reg_delta(looper,
                                                      txnPoolNodeSet,
                                                      sdk_pool_handle,
                                                      sdk_wallet_steward,
                                                      send_revoc_reg_entry_by_default):
    # We save timestamp of state changes.
    # looper and txnPoolNodeSet has "module" scope, therefore,
    # when we send request with FROM section, it's not a clean situation
    looper.runFor(3)
    # Assume, that send_revoc_reg_entry_by_default will add into revoked [1,2,3,4,5]
    rev_reg_req1, rev_reg_reply1 = send_revoc_reg_entry_by_default
    rev_reg_req1['operation'][VALUE][REVOKED] = []
    # Issue [1,2,3], Revoked now must be [4,5]
    rev_reg_req1['operation'][VALUE][ISSUED] = [1, 2, 3]
    rev_reg_req1['operation'][VALUE][PREV_ACCUM] = rev_reg_req1['operation'][VALUE][ACCUM]
    rev_reg_req1['operation'][VALUE][ACCUM] = randomString(10)
    rev_reg_req2, rev_reg_reply2 = sdk_send_and_check(
        [json.dumps(sdk_sign_request_from_dict(
            looper, sdk_wallet_steward, rev_reg_req1['operation']))],
        looper,
        txnPoolNodeSet,
        sdk_pool_handle)[0]
    # Revoke [10, 11]
    rev_reg_req2['operation'][VALUE][REVOKED] = [10, 11]
    rev_reg_req2['operation'][VALUE][ISSUED] = []
    rev_reg_req2['operation'][VALUE][PREV_ACCUM] = rev_reg_req2['operation'][VALUE][ACCUM]
    rev_reg_req2['operation'][VALUE][ACCUM] = randomString(10)
    rev_reg_req3, rev_reg_reply3 = sdk_send_and_check(
        [json.dumps(sdk_sign_request_from_dict(
            looper, sdk_wallet_steward, rev_reg_req2['operation']))],
        looper,
        txnPoolNodeSet,
        sdk_pool_handle)[0]
    reg_delta_req = build_get_revoc_reg_delta(looper, sdk_wallet_steward)
    reg_delta_req['operation'][REVOC_REG_DEF_ID] = rev_reg_req1['operation'][REVOC_REG_DEF_ID]
    reg_delta_req['operation'][FROM] = get_txn_time(rev_reg_reply1['result'])
    reg_delta_req['operation'][TO] = get_txn_time(rev_reg_reply3['result']) + 1000
    sdk_reply = sdk_send_and_check([json.dumps(reg_delta_req)], looper, txnPoolNodeSet, sdk_pool_handle)
    reply = sdk_reply[0][1]
    check_valid_proof(reply)


def test_state_proof_returned_for_get_revoc_reg_delta_with_only_to(
                                                      looper,
                                                      txnPoolNodeSet,
                                                      sdk_pool_handle,
                                                      sdk_wallet_steward,
                                                      send_revoc_reg_entry_by_default):
    rev_reg_req, rev_reg_reply = send_revoc_reg_entry_by_default
    reg_delta_req = build_get_revoc_reg_delta(looper, sdk_wallet_steward)
    del reg_delta_req['operation'][FROM]
    reg_delta_req['operation'][REVOC_REG_DEF_ID] = rev_reg_req['operation'][REVOC_REG_DEF_ID]
    reg_delta_req['operation'][TO] = get_utc_epoch() + 1000
    sdk_reply = sdk_send_and_check([json.dumps(reg_delta_req)], looper, txnPoolNodeSet, sdk_pool_handle)
    reply = sdk_reply[0][1]
    check_valid_proof(reply)


def test_state_proof_returned_for_delta_with_None_reply(
                                                      looper,
                                                      txnPoolNodeSet,
                                                      sdk_pool_handle,
                                                      sdk_wallet_steward,
                                                      send_revoc_reg_entry_by_default):
    rev_reg_req, rev_reg_reply = send_revoc_reg_entry_by_default
    reg_delta_req = build_get_revoc_reg_delta(looper, sdk_wallet_steward)
    del reg_delta_req['operation'][FROM]
    reg_delta_req['operation'][REVOC_REG_DEF_ID] = rev_reg_req['operation'][REVOC_REG_DEF_ID]
    reg_delta_req['operation'][TO] = get_utc_epoch() - 1000
    sdk_reply = sdk_send_and_check([json.dumps(reg_delta_req)], looper, txnPoolNodeSet, sdk_pool_handle)
    reply = sdk_reply[0][1]
    assert STATE_PROOF not in reply['result']


def test_state_proof_returned_for_delta_with_from_earlier(
                                                      looper,
                                                      txnPoolNodeSet,
                                                      sdk_pool_handle,
                                                      sdk_wallet_steward,
                                                      send_revoc_reg_entry_by_default):
    rev_reg_req, rev_reg_reply = send_revoc_reg_entry_by_default
    reg_delta_req = build_get_revoc_reg_delta(looper, sdk_wallet_steward)
    reg_delta_req['operation'][FROM] = get_utc_epoch() - 1000
    reg_delta_req['operation'][REVOC_REG_DEF_ID] = rev_reg_req['operation'][REVOC_REG_DEF_ID]
    reg_delta_req['operation'][TO] = get_utc_epoch() + 1000
    sdk_reply = sdk_send_and_check([json.dumps(reg_delta_req)], looper, txnPoolNodeSet, sdk_pool_handle)
    reply = sdk_reply[0][1]
    check_valid_proof(reply)



import pytest

from indy_node.test.auth_rule.auth_framework.claim_def import get_schema_json
from plenum.common.exceptions import RequestRejectedException

from indy_node.test.claim_def.test_send_claim_def import sdk_send_claim_def
from plenum.common.types import OPERATION

from indy_node.test.helper import build_auth_rule_request_json
from plenum.common.constants import TRUSTEE, STEWARD, DATA

from indy_common.authorize.auth_actions import ADD_PREFIX, EDIT_PREFIX
from indy_common.authorize.auth_constraints import AuthConstraint
from indy_common.constants import CLAIM_DEF

from plenum.test.helper import sdk_get_and_check_replies, sdk_sign_and_submit_op
from indy_node.test.auth_rule.helper import sdk_send_and_check_req_json


def send_and_check(looper, sdk_pool_handle, req_json, wallet):
    return sdk_send_and_check_req_json(
        looper, sdk_pool_handle, wallet, req_json)[0]


def test_auth_rule_transaction_for_edit(looper,
                                        txnPoolNodeSet,
                                        sdk_wallet_trustee,
                                        sdk_wallet_steward,
                                        sdk_pool_handle):
    constraint = AuthConstraint(role=STEWARD,
                                sig_count=1,
                                need_to_be_owner=False)
    req1 = build_auth_rule_request_json(
        looper, sdk_wallet_trustee[1],
        auth_action=ADD_PREFIX,
        auth_type=CLAIM_DEF,
        field='*',
        new_value='*',
        constraint=constraint.as_dict
    )

    send_and_check(looper, sdk_pool_handle, req1, wallet=sdk_wallet_trustee)

    constraint = AuthConstraint(role=TRUSTEE,
                                sig_count=1,
                                need_to_be_owner=False)
    req2 = build_auth_rule_request_json(
        looper, sdk_wallet_trustee[1],
        auth_action=EDIT_PREFIX,
        auth_type=CLAIM_DEF,
        field='*',
        old_value='*',
        new_value='*',
        constraint=constraint.as_dict
    )

    send_and_check(looper, sdk_pool_handle, req2, wallet=sdk_wallet_trustee)

    schema_json = get_schema_json(looper, sdk_pool_handle, sdk_wallet_trustee)

    reply = sdk_send_claim_def(looper, sdk_pool_handle, sdk_wallet_steward, 'tag_1', schema_json)

    req = reply[0][0]
    req[OPERATION][DATA]['primary']['n'] = 'abc'

    with pytest.raises(RequestRejectedException):
        resp = sdk_sign_and_submit_op(looper, sdk_pool_handle, sdk_wallet_steward, op=req[OPERATION])
        sdk_get_and_check_replies(looper, [resp])



import pytest
from indy_node.utils.node_control_utils import NodeControlUtil
# a -> bb, cc
# bb -> ddd
# cc -> eee
# ddd -> ffff, jjjj
# eee -> hhhh, iiii
# iiii -> gggg
pkg_a = ('a', '0.0.0')
pkg_bb = ('bb', '1.0.0')
pkg_cc = ('cc', '1.0.0')
pkg_ddd = ('ddd', '1.1.0')
pkg_eee = ('eee', '1.1.0')
pkg_ffff = ('ffff', '1.1.1')
pkg_jjjj = ('jjjj', '1.1.1')
pkg_hhhh = ('hhhh', '1.1.1')
pkg_iiii = ('iiii', '1.1.1')
pkg_gggg = ('gggg',)

mock_info = {
    "{}={}".format(*pkg_a): 'Version: {}\nDepends:{} (= {}), {} (= {})\n'.format(pkg_a[1], *pkg_bb, *pkg_cc),
    "{}={}".format(*pkg_bb): 'Version: {}\nDepends:{} (= {}), {}\n'.format(pkg_bb[1], *pkg_ddd, pkg_cc[0]),
    "{}={}".format(*pkg_cc): 'Version: {}\nDepends:{} (= {})\n'.format(pkg_cc[1], *pkg_eee),
    "{}={}".format(*pkg_ddd): 'Version: {}\nDepends:{} (= {}), {} (= {})\n'.format(pkg_ddd[1], *pkg_ffff, *pkg_jjjj),
    "{}={}".format(*pkg_eee): 'Version: {}\nDepends:{} (= {}), {} (= {})\n'.format(pkg_eee[1], *pkg_hhhh, *pkg_iiii),
    "{}={}".format(*pkg_ffff): 'Version: {}\nDepends: \n'.format(pkg_ffff[1]),
    "{}={}".format(*pkg_jjjj): 'Version: {}\nDepends: \n'.format(pkg_jjjj[1]),
    "{}={}".format(*pkg_hhhh): 'Version: {}\nDepends: \n'.format(pkg_hhhh[1]),
    "{}={}".format(*pkg_iiii): 'Version: {}\nDepends: {}\n'.format(pkg_iiii[1], pkg_gggg[0])
}


def mock_get_info_from_package_manager(*package):
    ret = ""
    for p in package:
        ret += mock_info.get(p, "")
    return ret


@pytest.fixture()
def patch_pkg_mgr(monkeypatch):
    monkeypatch.setattr(NodeControlUtil, '_get_info_from_package_manager',
                        lambda *x: mock_get_info_from_package_manager(*x))


@pytest.mark.parametrize("fltr_hld,res_dep",
                         [
                          ([pkg_a[0]], [pkg_a]),
                          ([pkg_a[0], pkg_cc[0]], [pkg_a, pkg_cc]),
                          ([pkg_a[0], pkg_jjjj[0]], [pkg_a, pkg_jjjj]),
                          ([pkg_a[0], pkg_bb[0], pkg_cc[0]], [pkg_a, pkg_bb, pkg_cc]),
                          ([pkg_cc[0]], [pkg_cc]),
                          ([pkg_a[0], pkg_bb[0], pkg_cc[0], pkg_ddd[0], pkg_ffff[0], pkg_jjjj[0]],
                           [pkg_a, pkg_bb, pkg_cc, pkg_ddd, pkg_ffff, pkg_jjjj]),
                          # Case, when gggg doesn't have version, dep just 'gggg'
                          ([pkg_gggg[0]], [pkg_gggg]),
                          ([pkg_a[0], pkg_bb[0], pkg_cc[0], pkg_ddd[0], pkg_eee[0], pkg_ffff[0], pkg_jjjj[0], pkg_iiii[0], pkg_hhhh[0], pkg_iiii[0]],
                           [pkg_a, pkg_bb, pkg_cc, pkg_ddd, pkg_eee, pkg_ffff, pkg_jjjj, pkg_iiii, pkg_hhhh])
                          ])
def test_deps_levels(patch_pkg_mgr, fltr_hld, res_dep):
    deps_list = NodeControlUtil.get_deps_tree_filtered('{}={}'.format(*pkg_a), hold_list=fltr_hld, deps_map={})
    flat_deps = []
    NodeControlUtil.dep_tree_traverse(deps_list, flat_deps)
    assert len(deps_list) == len(res_dep)
    for d in res_dep:
        assert "=".join(list(d)) in deps_list
